//-*-C++-*- Header file for the RawCD Class
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
//
// This file is unreadable, because the comment-docifier is
// brainedead.  Try a "make srcdocs" to generate the
// HTML-documentation.
//
#ifndef _RawCD_h_
#define _RawCD_h_
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <time.h>
#include "enums.h"


// This is the errno.h symbol that Linux returns when there is no
// media present in a removable media device.
#define NO_MEDIA_PRESENT ENXIO


/**
  This is the RawCD object.  It defines an API for cdrom-device
  access.
  
  The API is toolkit- and platform- independent.  Any
  platform-specific code will be in the *.C sources for each
  platform, which should include this file.
  
  The API uses standard Unix file operations, and requires a
  filename for the CDROM device.


  {\bf Private Members}

  There are numerous private member flags, variables, and methods.
  All [well, most] of the private methods are inlined.
  
  The private members are all specific to {\em this} implementation of
  a raw CD device.  If you decide to create your own implementation,
  it makes more sense to just keep this header file, but rewrite all
  of the code in the *.C file.  This is also what you should do if you
  need to port this class to systems other than Linux or to Linux SCSI
  drives.

  {\bf
  \begin{verbatim}
  
  ----------------------------------------------------------
  ----------------------------------------------------------
  \end{verbatim}
  }

  @memo Class for a raw CDROM device.

 */
class RawCD
{
   public:
      /**
	   The step by which to increase or decrease the CDROM drive
	   volume.  Initialized to 0.1.  [The volume ranges from 0 to 1.]
	   @memo	   
	   */
	 float volstep;

      /**
	   Class Constructor.
	   @memo	   
	   */
	 RawCD( void ) {
	   volstep = 0.1;
	   devfp = -1;
	   lrqpitem = 0;
	   noTray = false;
	   hasCloseIoctl = false;
	   isPlaying = isPaused = isStopped = isEjected = 
		playerError = playComplete = false;
	   isEmpty = true;	 
	 }

      /**
	   Overloaded Class Constructor.  Difference:  After initializing
	   the variables, it calls \Ref{InitDev}({\em devname}).  {\em
	   devname} is the filename of the CDROM device.
	   @memo	   
	   */
	 RawCD( char *devname ) {
	   volstep = 0.1;
	   devfp = -1;
	   lrqpitem = 0;
	   noTray = false;
	   hasCloseIoctl = false;
	   InitDev(devname);
	 }

      /**
	   Overloaded version.  Sets an internal pointer to {\em
	   devname}, and closes the CDROM device if it's currently open.
	   It then calls \Ref{InitDev}(void) to re-initialize the device
	   using the new device filename.
	   @memo	   
	   */
	 // Overloaded version.
	 void InitDev( char *devname ) {
	   device_file_name = devname;
	   if(devfp != -1)
		 {
		 close(devfp);
		 devfp = -1;
		 }
	   InitDev();
	 }

      /**
	   Sets how a RawCD object controls the CDROM tray.  Use {\em
	   close_ioctl}=true if you can close the CDROM tray via an ioctl
	   call.  Use {\em trayless}=true if the CDROM has no retractable
	   tray.
	   @memo */
	 void TrayControl( bool close_ioctl=false, bool trayless=false ) {
	    hasCloseIoctl = close_ioctl;
	    noTray = trayless;
	 }

      /**
	   You should call \Ref{UpdateStatus}() before calling this
	   method.

	   @return The current volume of the CDROM device.
	   @memo	   
	   */
	 float Volume(void) {
	   return vol;
	 }

      /**
	   Returns the position on the CD of the {\em idx}-th track, in
	   units of frames.
	   
	   {\em idx}=0 refers to the first item.
	   @memo	   
	   */
	 int ToC_Entry(int idx) {
	   idx += first_item;
	   return toc[idx];
	 }

      /**
	   Sets the lrqp-counter to {\em idx}.  
	   {\em idx}=0 refers to the first item.

	   The lrqp-counter is used by \Ref{PlayItem}() for synchronization
	   purposes.

	   @see PlayItem
	   @memo	   
	   */
	 void lrqpSet(int idx) {
	    lrqpitem = idx + first_item;
	 }
	    
      /**
	   Returns the length, in units of frames, of the {\em idx}-th
	   track.

	   {\em idx}=0 refers to the first item.
	   @memo	   
	   */
	 int ItemLength(int idx) {
	   idx += first_item;
	   if (idx < 1)
		 return toc[1]-toc[0];
	   else if (idx > last_item)
		 idx = last_item;
	   return toc[idx+1] - toc[idx];
	 }

      /**
	   You should call \Ref{UpdateStatus}() before calling this
	   method.

	   @return A \Ref{CdStatus} {\bf enum} containing the current
	   status of the device.
	   @memo	   
	   */
	 CdStatus getStatus(void) {
	   return (CdStatus)( ((int)isPlaying*cdsPlaying) |
					  ((int)isPaused*cdsPaused) |
					  ((int)isStopped*cdsStopped) |
					  ((int)isEjected*cdsEjected) |
					  ((int)isEmpty*cdsEmpty) |
					  ((int)playerError*cdsError) |
					  ((int)playComplete*cdsComplete)
		);
	 }

      /**
	   You should call \Ref{UpdateStatus}() before calling this
	   method.

	   @return The index of the current track playing.  A value of 0
	   refers to the first item.
	   @memo	   
	   */
	 int CurrentItemIndex(void) {
	   return citem-first_item;
	 }

      /**
	   You should call \Ref{UpdateStatus}() before calling this
	   method.

	   @return The length, in frames, of the current item playing.
	   @memo	   
	   */
	 int CurrentItemLength(void) {
	   return (toc[citem+1]-toc[citem]);
	 }

      /**
	   You should call \Ref{UpdateStatus}() before calling this
	   method.

	   @return The time elapsed, relative to the current item
	   playing.
	   @memo	   
	   */
	 int CurrentItemTime(void) {
	   return rel_ctime;
	 }

      /**
	   @return The number of playable tracks on the CD [excludes the
	   leadout track and the nonexistent null-track].
	   @memo	   
	   */
	 int ToC_Length(void) {
	   return nitems;
	 }

      /**
	   Uses the device filename [set by its overloaded version or by the
	   constructor]  to open the CDROM device, then loads the ToC, and
	   lastly calls \Ref{UpdateStatus}().
	   @memo	   
	   */
	 void InitDev(void);

      /**
	   Increments the volume by a step.
	   @memo	   
	   */
	 void IncVol(void);

      /**
	   Decrements the volume by a step.
	   @memo	   
	   */
	 void DecVol(void);

      /**
	   Stops play, updating the device status and setting internal flags.
	   Only does something if the drive is closed and has a disc.
	   @memo	   
	   */
	 void Stop(void);

      /**
	   Toggles the pause state of the cdrom.  Only does something if the
	   drive is playing.
	   @memo	   
	   */
	 void PauseResume(void);

      /**
	   Play track {\em itemno} between two blocks of time, starting at
	   {\em t0} and ending at {\em t_f}.  {\em t0} and {\em t_f} are
	   relative to the start of the specified track.  {\em itemno}=0
	   corresponds to the first playable track.
	   
	   If {\em t_f}<={\em t0}, {\bf PlayItemBlock} returns an error.  If
	   {\em t_f} is 0, negative, or beyond the end of the CD, {\bf
	   PlayItemBlock} simply plays to the end of the disc.
	   
	   The flag {\em reposition} is used to control what {\bf
	   PlayItemBlock} does if the drive is already playing.  If {\em
	   reposition=true}, the drive stops and restarts at the new
	   location.  Otherwise, {\bf PlayItemBlock} does nothing.
	   @memo	   
	   */
	 void PlayItemBlock(int itemno, int t0, int t_f=-1,
					bool reposition=true);

      /**
	   Play track {\em itemno}.  {\em itemno}=0 corresponds to the first
	   playable track.
	   
	   The flag {\em reposition} is used to control what {\bf
	   PlayItem} does if the drive is already playing.  If {\em
	   reposition=true}, the drive stops and restarts at the new
	   location.  Otherwise, {\bf PlayItem} does nothing.
	   
	   This is where the lrpq-index comes in.  It specifies the index of
	   the last specifically-requested play item, and is set by both {\bf
	   PlayItem} and \Ref{PlayItemBlock}.  If:
	   \begin{enumerate}
	   \item {\em reposition=true}
	   \item the requested track is the one immediately following the last
	   one requested [{\it i.e. itemno=lrqp+1}].
	   \item the current play position is +/-1 second of the beginning of
	   {\em itemno}
	   \end{enumerate}
	   ...then {\bf PlayItem} just updates some counters and returns,
	   allowing the drive to keep playing.  If the last two conditions
	   aren't met, then the drive stops an restarts.
	   
	   The idea behind this elaborate scheme is to keep play smooth.  As
	   long as your CDROM drive keeps playing through, in sequence, once
	   it's started playing, this scheme will work.
	   @memo	   
	   */
	 void PlayItem(int itemno, bool reposition=false);

      /**
	   Update the status flags and counter variables.  Does nothing if the
	   tray is ejected or empty.
	   @memo	   
	   */
	 void UpdateStatus(void);

      /**
	   Ejects the CD.
	   @memo	   
	   */
	 void Eject(void);

      /**
	   Closes the tray by one of two methods, depending on how the
	   routine is compiled.  It can use the CDROMCLOSETRAY {\tt ioctl}, or
	   it can simply close and reopen the CDROM device.  It always calls
	   \Ref{InitDev}(void) after everything else. 
	   @memo	   
	   */
	 void CloseTray(void);

	 
   private:
	 //
	 float vol;
	 int citem,rel_ctime,abs_ctime,lrqpitem;
	 int *toc,nitems,first_item,last_item;
	 int devfp;
	 char *device_file_name;
	 bool isPlaying,isPaused,isStopped,isEjected,isEmpty;
	 bool playerError,playComplete;
	 bool noTray,hasCloseIoctl;
	 inline int msf2int(struct cdrom_msf0);
	 inline struct cdrom_msf0 int2msf(int tt);
	 inline struct cdrom_msf int2msf_block(int t0, int tf);
	 inline bool ioctlOk(int stat);
};

// Okay to define the inlined private functions in the source file,
// since that's the only place they'll ever get used from.

#endif // _RawCD_h_
//||||||||||||||||||||||||||||||||
//
// End
// 
